"""Command helper functions."""

# Absolute import (the default in a future Python release) resolves
# the logging import as the Python standard logging module rather
# than qiutil.logging.
from __future__ import absolute_import
import os
import tempfile
from logging import (ERROR, DEBUG)
from .logging import configure


def add_options(parser):
    """
    Adds the standard ``--log``, ``--quiet``, ``--verbose`` and ``--debug``
    options to the given command line argugment parser.
    """
    parser.add_argument('-l', '--log', help='the log file', metavar='FILE')
    verbosity_grp = parser.add_mutually_exclusive_group()
    verbosity_grp.add_argument(
        '-q', '--quiet', help="only log error messages", dest='log_level',
        action='store_const', const=ERROR)
    verbosity_grp.add_argument(
        '-d', '--debug', help='log debug messages', dest='log_level',
        action='store_const', const=DEBUG)


def configure_log(app, opts):
    """
    Configures the logger.

    :param app: the application name
    :param opts: the following keyword options:
    :keyword log: the log file
    :keyword log_level: the log level
    """
    log_cfg = {}
    if 'log' in opts:
        log_file = os.path.abspath(opts.get('log'))
        log_cfg['filename'] = log_file
    if 'log_level' in opts:
        log_cfg['level'] = opts.get('log_level')
    configure(app, **log_cfg)



import os


def which(program):
    """
    Returns the file path of the first executable of the given program
    name in the system ``PATH`` environment variable.
    
    This function is a system-independent Python equivalent of the \*nix
    ``which`` command.
    
    :Note: The implementation is adapted from
        http://stackoverflow.com/questions/377017/test-if-executable-exists-in-python.
    
    :param program: the program name to check
    :return: the filename in the system path
    """
    fpath, fname = os.path.split(program)
    if fpath:
        if _is_executable(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if _is_executable(exe_file):
                return exe_file

    return None


def _is_executable(fpath):
    """
    :param fpath: the file path to check
    :return: whether the file exists and is executable
    """
    return os.path.isfile(fpath) and os.access(fpath, os.X_OK)



import __builtin__
import os
import re
import inspect
import gzip
from contextlib import contextmanager
from .uid import generate_string_uid

SPLITEXT_PAT = re.compile("""
    (.*?)           # The file path without the extension
    (               # The extension group
        (\.\w+)+    # The (possibly composite) extension
    )?              # The extension is optional
    $               # Anchor to the end of the file path
    """, re.VERBOSE)
"""
Regexp pattern that splits the name and extension.

:see: :meth:`splitexts`
"""


@contextmanager
def open(filename):
    """
    Opens the given file. If the file extension ends in ``.gz``,
    then the content is uncompressed.

    :param filename: the file path
    :return: the file input stream
    :raise: IOError if the file cannot be read
    """
    _, ext = os.path.splitext(filename)
    # The gzip or normal file open context function.
    if ext == '.gz':
        context = gzip.open
    else:
        context = __builtin__.open

    # Open the file.
    with context(filename) as fp:
        yield fp


def splitexts(path):
    """
    Splits the given file path into a name and extension.
    Unlike ``os.path.splitext``, the resulting extension can be composite.
    
    Example:
    
    >>> import os
    >>> os.path.splitext('/tmp/foo.nii.gz')
    ('/tmp/foo.nii', '.gz')
    >>> from qiutil.file import splitexts
    >>> splitexts('/tmp/foo.3/bar.nii.gz')
    ('/tmp/foo.3/bar', '.nii.gz')
    
    :param path: the file path
    :return: the *(prefix, extensions)* tuple
    """
    matches = SPLITEXT_PAT.match(path).groups()[:2]
    # Pad with an empty extension, if necessary.
    matches += (None,) * (2 - len(matches))
    
    return tuple(matches)


def generate_file_name(ext=None):
    """
    Makes a valid file name which is unique to within one microsecond of calling
    this function. The file name is ten characters long without the extension.
    
    :param ext: the optional file extension, with leading period delimiter
    :return: the file name
    """
    fname = generate_string_uid()
    if ext:
        return fname + ext
    else:
        return fname


class FileIterator(object):
    """
    FileIterator is a generator class which iterates over the files contained
    recursively in the :meth:`__init__` file specifications.
    """
    def __init__(self, *filespecs):
        """
        :param filespecs: the files, directories or file generators over which to iterate
        """
        self._filespecs = filespecs

    def __iter__(self):
        return self.next()

    def next(self):
        """
        Iterates over the files as follows:
        
        - If the current file specification is a file object, then yield that
          file object.
        
        - If the current file specification is a non-directory file path,
          then yield that file path.
        
        - If the current file specification is a directory, then yield each file
          contained recursively in that directory.
        
        - If the current file specification is a generator, then yield each
          generated item
        
        :yield: the next file object or path
        :raise IOError: if an iterated file path is neither a directory nor file
        :raise ValueError: if an iterated item is not a file object, path
            or generator
        """
        for spec in self._filespecs:
            if isinstance(spec, file):
                yield spec
            elif isinstance(spec, str):
                if os.path.isfile(spec):
                    yield spec
                elif os.path.isdir(spec):
                    for root, dirs, fnames in os.walk(spec):
                        for fn in fnames:
                            yield os.path.join(root, fn)
                else:
                    raise IOError("File not found: %s" % spec)
            elif inspect.isgenerator(spec):
                for f in spec:
                    yield f
            else:
                raise ValueError("File iteration item is not supported:"
                                 " %s" % spec.__class__)



import time
import calendar
import struct
import base64


def generate_uid(yr=None):
  
    """
    Makes an id that is unique to within one microsecond of calling
    this function for the execution context which runs this method.
    
    This is a simple, fast id generator that is adequate for most
    purposes.
    
    The base year parameter should be no greater than the current
    year and should not change over the lifetime of all objects
    created using that year. A higher base year results in a smaller
    UID.
    
    :param: the optional base year for the time offset (default 2001)
    :rtype: long
    :return: the UID
    """
    # The default year.
    if not yr:
      yr = 2001
    # A starting time prior to now.
    start = time.mktime(calendar.datetime.date(yr, 01, 01).timetuple())
    # A long which is unique to within one microsecond.
    return long((time.time() - start) * 1000000)


def generate_string_uid(yr=None):
    """
    Makes a string id based on the :meth:`generate_uid` value.
    The string id is a URL-safe base64-encoded string without trailing
    filler or linebreak. It is thus suitable for file names as well as
    URLs.
    
    The generated id is ten characters long for the default base year.
    
    :param yr: the optional :meth:`generate_uid` base year parameter
    :rtype: str
    :return: the UID
    """
    # The long uid.
    uid = generate_uid(yr)
    # Encode the long uid without trailing filler or linebreak.
    return base64.urlsafe_b64encode(struct.pack('L', uid)).rstrip('A=\n')



# Absolute import (the default in a future Python release) resolves
# the collections import as the Python standard collections module
# rather than this module of the same name.
from __future__ import absolute_import
from copy import copy
from collections import (Iterable, Mapping, defaultdict)
import functools

def is_nonstring_iterable(obj):
    """
    :param obj: the object to check
    :return: whether the given object is a non-string iterable object
    """
    return isinstance(obj, Iterable) and not isinstance(obj, str)


def to_series(items, conjunction='and'):
    """
    Formats the given items as a series string.

    Example:

    >>> to_series([1, 2, 3])
    '1, 2 and 3'

    :param items: the items to format in a series
    :param conjunction: the series conjunction
    :return: the items series
    :rtype: str
    """
    if not items:
        return ''
    prefix = ', '.join([str(i) for i in items[:-1]])
    suffix = str(items[-1])
    if not prefix:
        return suffix
    else:
        return (' ' + conjunction + ' ').join([prefix, suffix])

def nested_defaultdict(factory, levels=0):
    """
    Makes a defaultdict for the given factory and number of levels, e.g.::

        >> from qiutil.collections import nested_defaultdict as dd
        >> dd(list, 0)[1]
        []
        >> dd(dict, 2)[1][2][3]
        {}

    Note that the default levels parameter value 0 is synonymous with the
    standard Python collections defaultdict, i.e.::

        dd(list)

    is the same as::

        dd(list, 0)

    or::

        from collections import defaultdict
        defaultdict(list)
    
    Thus, this ``nested_defaultdict`` function can serve as a drop-in
    replacement for ``defaultdict``.

    :param factory: the 0th level defaultdict factory.
    :param levels: the number of levels
    """
    # The recursive nested dictionary generator, where f is the factory
    # and n is the number of levels.
    dd = lambda f, n: defaultdict((lambda: dd(f, n - 1)) if n else f)

    return dd(factory, levels)


def update(target, *sources, **opts):
    """
    Updates the given target object from the given source objects.
    The target object can be a dictionary, list or set. The target
    and sources are validated for compatibility as follows:
    
    * If the target object is a Mapping, then the sources must
      be Mappings.
    
    * Otherwise, if the target object is a list or set, then the
      sources must be non-string iterables.
    
    The target is updated from the sources in order as follows:
    
    * If the target object is a Mapping and the *recursive* flag is
      falsey, then the standard Python dictionary update is applied.
    
    * If the target object is a Mapping and the *recursive* flag is
      truthy, then the update is applied recursively to nested
      dictionaries, e.g.:
      
      >> from qiutil.collections import update
      >> target = dict(a=dict(aa=1))
      >> update(target, dict(a=dict(aa=2, ab=3)))
      >> target
      {'a': {'aa': 2, 'ab': 3}}
    
    * If the target object is a list or set, then the source items
      which are not yet in the target are added to the target, e.g.:
    
      >> from qiutil.collections import update
      >> target = [1, 2, 2, 5]
      >> update(target, [4, 2, 6, 6])
      >> target
      [1, 2, 2, 5, 4, 6, 6]
    
    This function adapts the solution offered in a
    `StackOverflow post <http://stackoverflow.com/questions/3232943/update-value-of-a-nested-dictionary-of-varying-depth>`
    to support lists, sets and multiple sources.
    
    :param target: the dictionary to update
    :param sources: the update source dictionaries
    :param opts: the following keyword options:
    :keyword recursive: if True, then apply the update recursively to
        nested dictionaries
    """
    # Validate the sources.
    _validate_update_compatibility(target, *sources)
    # Make the update helper function. This idiom refactors the source
    # iteration block into a callable function with a sole source argument.
    # This pattern is a little obscure to those not well-versed in functional
    # programming, but it is cleaner than the alternative of embedding the
    # _updater logic into the source iteration.
    updater = _create_updater(target, **opts)
    # Apply the successive source updates.
    for source in sources:
        updater(source)


def _create_updater(target, **opts):
    """
    :param target: the update target
    :param opts: the following keyword options:
    :keyword recursive: if True, then apply the update recursively to
        nested dictionaries
    :return: the function to apply to a *source* argument
    """
    if isinstance(target, Mapping):
        if opts.get('recursive'):
            return functools.partial(_update_dict_recursive, target)
        else:
            # Apply the standard Python dictionary update.
            return lambda src: target.update(src)
    else:
        return functools.partial(_update_collection, target)


def _update_dict_recursive(target, source):
    for key, srcval in source.iteritems():
        if key in target:
            tgtval = target[key]
            # If the target value can be merged from the source
            # value, then replace the target value with a shallow
            # copy and update it recursively.
            if isinstance(tgtval, Mapping) and isinstance(srcval, Mapping):
                target[key] = tgtval = copy(tgtval)
                _update_dict_recursive(tgtval, srcval)
                continue
        # Set the target item.
        target[key] = copy(srcval)


def _validate_update_compatibility(target, *sources):
    if isinstance(target, Mapping):
        for source in sources:
            if not isinstance(source, Mapping):
                raise TypeError("Update source is incompatible with the"
                                " dictionary target: %s" % source)
    elif isinstance(target, list) or isinstance(target, set):
        for source in sources:
            if not is_nonstring_iterable(source):
                raise TypeError("Update source is incompatible with the"
                                " collection target: %s" % source)
    else:
        raise TypeError("Update target is type is not supported: %s" % target)


def _update_collection(target, source):
    """
    Adds to the target those source items which are not
    yet in the target, as described in :meth:`update`.
    
    :param target: the list or set to update
    :param source: the input non-string iterable
    :raise TypeError: if the target is neither a list or set
    """
    if isinstance(target, set):
        target.update(source)
    elif isinstance(target, list):
        exclude = set(target)
        diff = (item for item in source if item not in exclude)
        target.extend(diff)
    else:
        raise TypeError("Update target type not supported")


class ImmutableDict(dict):

    """
    ImmutableDict is a dictionary that cannot be changed after creation.

    An ImmutableDict is *not* hashable and therefore cannot be used as a
    dictionary key or set member. See http://www.python.org/dev/peps/pep-0351
    for the rationale.
    """

    def __init__(self, *args, **kwargs):
        super(ImmutableDict, self).__init__(*args, **kwargs)

    def __setitem__(self, key, value):
        """
        :raise NotImplementedError: always
        """
        raise NotImplementedError("The dictionary is immutable: %s" % self)


EMPTY_DICT = ImmutableDict()
"""
An immutable empty dictionary.
This constant serves as an efficient method return default value.
"""



"""The Quantitative Imaging utility module."""

__version__ = '2.1.8'
"""
The one-based major.minor.patch version.
The version numbering scheme loosely follows http://semver.org/.
The major version is incremented when a significant feature
set is introduced. The minor version is incremented when there
is a functionality change. The patch version is incremented when
there is a refactoring or bug fix. All major, minor and patch
version numbers begin at 1.
"""

# Import collections, file and logging, since these are also
# standard Python libraries. This import allows the client to
# use the nested modules directly, e.g.:
#   with qiutil.file.open(...):
# rather than:
#   from qiutil import file
#   with file.open(...): # Misleading
#  or:
#   from qiutil import file as qifile
#   with qifile.open(...): # Awkward
from . import (collections, file, logging) 



import os
from contextlib import contextmanager

@contextmanager
def cd(path):
    """
    Safely changes directory in a ``with`` context, e.g.::
    
        from qiutil.cd import cd
        with cd('/path/to/dir'):
            # Do something...  
    
    This :meth:`cd` function is functionally equivalent to the
    `Grizzled <http://software.clapper.org/grizzled-python/>` 
    ``working_directory`` function. It is included in this
    ``qiutil`` package for convenience.
    
    :Note: this :meth:`cd` function is intended for use only
    within a context. Use outside of a context does not change
    the working directory, e.g.:
    
    >> import os
    >> from qiutil.cd import cd
    >> wd = os.getcwd()
    >> cd('/tmp')
    >> os.getcwd() == wd
    True
    
    :param path: the directory path to change to
    """
    prevdir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prevdir)



import os
import re
import subprocess
import glob
from setuptools import (setup, find_packages)

def version(package):
    """
    Return package version as listed in the `__init.py__` `__version__`
    variable.
    """
    init_py = open(os.path.join(package, '__init__.py')).read()
    return re.search("__version__ = ['\"]([^'\"]+)['\"]", init_py).group(1)


def requires():
    """
    @return: the ``requirements.txt`` package specifications
    """
    with open('requirements.txt') as f:
        return f.read().splitlines()
        

def readme():
    with open("README.rst") as f:
        return f.read()

# Note: bdist binary distribution is not supported because of the
# following setuptools bug:
#
# * setuptools bdist cannot include selective top-level files and
#   directories. Contrary to
#   http://stackoverflow.com/questions/7522250/how-to-include-package-data-with-setuptools-distribute,
#   there is no known combination of package_data, include_package_data
#   and MANIFEST.in that includes selective top-level files and directories
#   in both the source and binary distributions.
#
#   The work-around is to only build the source distribution as follows:
#
#   1. Set MANIFEST.in to the following:
#
#          include *.rst *.txt
#          recursive-exclude test *
#
#      recursive exclude excludes the test files. Note that this entry is
#      only strictly necessary for some projects, e.g. qidicom, but not
#      others, e.g. qiutil. The reason for this difference is unknown, but
#      is probably due to another obscure setuptools bug. At any rate, it
#      doesn't hurt to include it in all projects. It is advisable to
#      exclude the test directory because only python files are included,
#      and the python test cases often require test fixtures.
#
#   2. Move top-level directories to include in the build, e.g. conf,
#      to qiutil.
#
#   3. Add the package_data option to the setup call, e.g.:
#
#          package_data = dict(qiutil=['conf/*']),
#  
#   4. Build the source distribution:
#
#          python setup.py sdist
#
#   5. Upload the distribution to PyPI:
#
#          twine upload dist/*
#
# TODO - revisit this when the python 3.x package module is released.
#
setup(
    name = 'qiutil',
    version = version('qiutil'),
    author = 'OHSU Knight Cancer Institute',
    author_email = 'loneyf@ohsu.edu',
    platforms = 'Any',
    license = 'MIT',
    keywords = 'Imaging QIN',
    packages = find_packages(exclude=['test**']),
    package_data = dict(qiutil=['conf/*']),
    url = 'https://github.com/ohsu-qin/qiutil',
    description = 'Quantitative Imaging helper utilities.',
    long_description = readme(),
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ],
    install_requires = requires()
)



import os
import re
import ast
from ConfigParser import ConfigParser as Config
from .collections import to_series

from .logging import logger


def read_config(*filenames):
    """
    Reads and parses the given configuration file.
    
    :param filenames: the input configuration file names
    :return: the configuration
    :rtype: :class:`qiutil.ast_config.ASTConfig`
    :raise IOError: if no configuration files could not be read
    """
    cfg = ASTConfig()
    filenames = [os.path.abspath(fname) for fname in filenames]
    cfg_files = cfg.read(filenames)
    if not cfg_files:
        raise ValueError(
            "Configuration file could not be read: %s" % filenames)
    logger(__name__).debug("Loaded configuration from %s with sections %s." %
          (to_series(cfg_files), to_series(cfg.sections())))

    return cfg


class ASTConfig(Config):

    """
    ASTConfig parses a configuration option values as AST objects.
    Strings are quoted, if necessary.
    A bracked value is parsed as a list.
    A case-insensitive match on ``true`` or ``false`` is parsed
    as ``True``, resp. ``False``.
    
    For example, given the configuration file ``tuning.cfg`` with content::
        
        [Tuning]
        method = FFT
        iterations = [[1, 2], 5]
        parameters = [(1,), (2, 3), -0.4]
        two_tailed = false
        threshold = 4.0
        plugin_args = {'qsub_args': '-pe mpi 48-120'}
    
    then:
    
    >>> cfg = ASTConfig('tuning.cfg')
    >>> cfg['Tuning']
    {'method': u'FFT', 'parameters' = [(1,), (2, 3), -0.4],
     'iterations': [[1, 2], 5],
     'two_tailed': False, 'threshold': 4.0,
     'plugin_args': {'qsub_args': '-pe mpi 48-120'}}
    """

    BUNCH_PAT = """
        \%(left)s   # The left delimiter
        (.*)        # The list items
        \%(right)s$ # The right delimiter
        """
    """A bunch string pattern."""

    LIST_PAT = re.compile(BUNCH_PAT % dict(left='[', right=']'), re.VERBOSE)
    """A list string pattern."""

    TUPLE_PAT = re.compile(BUNCH_PAT % dict(left='(', right=')'), re.VERBOSE)
    """A tuple string pattern."""

    DICT_PAT = re.compile(BUNCH_PAT % dict(left='{', right='}'), re.VERBOSE)
    """A dictionary string pattern."""

    EMBEDDED_BUNCH_PAT = """
        ([^%(left)s]*)          # A prefix without the left delimiter
        (\%(left)s.*\%(right)s)? # The embedded item
        ([^%(right)s]*)         # A suffix without the right delimiter
        $                       # The end of the value
    """

    EMBEDDED_LIST_PAT = re.compile(EMBEDDED_BUNCH_PAT %
                                   dict(left='[', right=']'), re.VERBOSE)
    """A (prefix)(list)(suffix) recognition pattern."""

    EMBEDDED_TUPLE_PAT = re.compile(EMBEDDED_BUNCH_PAT %
                                    dict(left='(', right=')'), re.VERBOSE)
    """A (prefix)(tuple)(suffix) recognition pattern."""

    EMBEDDED_DICT_PAT = re.compile(EMBEDDED_BUNCH_PAT %
                                   dict(left='{', right='}'), re.VERBOSE)
    """A (prefix)(dictionary)(suffix) recognition pattern."""

    PARSEABLE_ITEM_PAT = re.compile("""
        (
            True              # The True literal
            | False           # The False literal
            | -?\.\d+         # A decimal with leading period
            | -?\d+(\.\d*)?   # A number
            | -?\d+(\.\d*)?e[+-]\d+ # A floating point
            | \'.*\'          # A single-quoted string
            | \".*\"          # A double-quoted string
        )$                    # The end of the value
        """, re.VERBOSE)
    """A non-list string parseable by AST."""

    def __iter__(self):
        return self.next()

    def next(self):
        """
        :yield: the next *(section, {item: value})* tuple
        """
        for s in self.sections():
            yield (s, self[s])

    def __contains__(self, section):
        """
        :param: the config section to find
        :return: whether this config has the given section
        """
        return self.has_section(section)

    def __getitem__(self, section):
        """
        :param section: the configuration section name
        :return: the section option => value dictionary
        """
        return {name: self._parse_entry(name, value)
                for name, value in self.items(section)}

    def _parse_entry(self, name, s):
        """
        :param name: the option name
        :param s: the option string value to parse
        :return: the parsed AST value
        :raise SyntaxError: if the value cannot be parsed
        """
        if s:
            ast_value = self._to_ast(s)
            try:
                return ast.literal_eval(ast_value)
            except Exception:
                logger(__name__).error("Cannot load the configuration entry"
                                       " %s: %s parsed as %s" % (name, s, ast_value))
                raise

    def _to_ast(self, s):
        """
        :param s: the input string
        :return: the equivalent AST string
        """
        # Trivial case.
        if not s:
            return
        # If the input is a boolean, number or already quoted,
        # then we are done.
        if ASTConfig.PARSEABLE_ITEM_PAT.match(s):
            return s

        # If the string is a list, then make a quoted list.
        # Otherwise, if the string signifies a boolean, then return the boolean.
        # Otherwise, quote the content.
        if self._is_bunch(s):
            return self._quote_bunch(s)
        elif s.lower() == 'true':
            return 'True'
        elif s.lower() == 'false':
            return 'False'
        elif s.lower() in ['null', 'none', 'nil']:
            return 'None'
        else:
            return '"%s"' % s.replace('"', '\\"')

    def _is_bunch(self, s):
        return (ASTConfig.LIST_PAT.match(s) or
                ASTConfig.TUPLE_PAT.match(s) or
                ASTConfig.DICT_PAT.match(s))

    def _quote_bunch(self, s):
        # A dictionary must already be properly quoted.
        if s[0] == '{':
            return s
        quoted_items = self._quote_bunch_content(s[1:-1])
        if len(quoted_items) == 1:
            quoted_items.append('')
        return s[0] + ', '.join(quoted_items) + s[-1]

    def _quote_bunch_content(self, s):
        """
        :param s: the comma-separated items
        :return: the bunch of quoted items
        """
        pre, mid, post = ASTConfig.EMBEDDED_LIST_PAT.match(s).groups()
        if not mid:
            pre, mid, post = ASTConfig.EMBEDDED_TUPLE_PAT.match(s).groups()
        if not mid:
            pre, mid, post = ASTConfig.EMBEDDED_DICT_PAT.match(s).groups()
        if mid:
            items = []
            if pre:
                items.extend(self._quote_bunch_content(pre))
            # Balance the left and right delimiter.
            left = mid[0]
            right = mid[-1]
            count = 1
            i = 1
            while (count > 0):
                if mid[i] == left:
                    count = count + 1
                elif mid[i] == right:
                    count = count - 1
                i = i + 1
            post = mid[i:] + post
            mid = mid[0:i]
            items.append(self._quote_bunch(mid))
            if post:
                items.extend(self._quote_bunch_content(post))
            return items
        else:
            # No embedded bunch.
            items = re.split('\s*,\s*', s)
            quoted_items = [self._to_ast(item) for item in items if item]
            return quoted_items



from .collections import is_nonstring_iterable


def on(root):
    """
    :param root: the nested dictionary to wrap
    :return: a new DictionaryHierarchy on the given root dictionary
    """
    return DictionaryHierarchy(root)


class DictionaryHierarchy(object):

    """
    A DictionaryHierarchy wraps a nested dictionary. The DictionaryHierarchy iterator
    enumerates the root -> leaf paths.
    
    For example, the hierarchy of a nested dictionary given by::
    
        1 : 'a'
        2 : 
          3 : 'b'
          4 : 'c', 'd'
          6 :
            7 : 'e'
        8 : 'f'
    
    results in the following paths::
    
        1, 'a'
        2, 3, 'b'
        2, 4, 'c'
        2, 4, 'd'
        2, 6, 7, 'e'
        8, 'f'
    """

    def __init__(self, root):
        """
        :param root: the nested dictionary to wrap by this hierarchy
        :type root: dict
        :raise ArgumentError: if the given root is not a dictionary 
        """
        if not isinstance(root, dict):
            raise TypeError(
                "The dictionary hierarchy root is not a dictionary: {1}" % root)
        self.root = root

    def __iter__(self):
        return self.Iterator(self.root)

    class Iterator(object):

        def __init__(self, root):
            """
            :param root: the nested dictionary to iterate over
            :type root: dict
            :raise ArgumentError: if the given root is not a dictionary 
            """
            self.base = root.iteritems()
            self.child = self.path = None

        def __iter__(self):
            return self

        def next(self):
            """
            Returns the next root-to-leaf path, determined as follows:
            
            - The first path member is the the current key in the wrapped root (key, value) iteration.
            
            - If the value is a dictionary, then recursively iterate over path tails given by that
              dictionary's hierarchy.
            
            - Otherwise, if the value is a non-string iterable, then recursively iterate over
              path tails given by that value's iteration.
            
            - Otherwise, the last member in the path is the value.
        
            :yield: the next path list
            """
            if self.child:
                try:
                    # Try to advance the child and replace the path tail.
                    self.path = self.path[:1] + self.child.next()
                    return self.path
                except StopIteration:
                    # Not an error; iterate to the next child.
                    self.child = None

            # Iterate to the next (key, value) pair.
            key, value = self.base.next()
            # The path begins with the key.
            self.path = [key]
            # If the next value is a non-string collection, then iterate to the child.
            # Otherwise, append the value to the path.
            if is_nonstring_iterable(value):
                # Wrap a dictionary value as a child hierarchy.
                if isinstance(value, dict):
                    self.child = iter(DictionaryHierarchy(value))
                else:
                    self.child = iter([[item] for item in value])
                # Recursively iterate into the new child.
                return self.next()
            else:
                self.child = None
                # The path is the (key, value) pair.
                self.path.append(value)
                return self.path



# Absolute import (the default in a future Python release) resolves
# the logging import as the Python standard logging module rather
# than this module of the same name.
from __future__ import absolute_import
import os
import logging
import logging.config
import yaml
from . import collections as qicollections


def logger(name):
    """
    This method is the preferred way to obtain a logger.

    Example:

    >>> from qiutil.logging import logger
    >>> logger(__name__).debug("Starting my application...")

    :param name: the caller's context ``__name__``
    :return: the Python Logger instance
    """
    # Configure on demand.
    if not hasattr(logger, 'configured'):
        configure(name)

    return logging.getLogger(name)


def configure(app, cfg_file=None, **opts):
    """
    Configures the global logger. The logging configuration is obtained from
    from the given keyword arguments and the YAML_ logging configuration files.
    The following logging configuration files are loaded in low-to-high
    precedence:

    - the ``qiutil`` module ``conf/logging.yaml`` file

    - the ``logging.yaml`` file in the current directory

    - the file specified by the ``LOG_CFG`` environment variable

    - the *cfg_file* parameter

    The ``opts`` keyword arguments specify simple logging parameters that
    override the configuration file settings. The keyword arguments
    can include the *filename* and *level* short-cuts, which are handled
    as follows:

    - if the *filename* is None, then file logging is disabled. Otherwise,
      the file handler file name is set to the *filename* value.

    - The *level* is set for the ``application`` logger. In addition, if
      the ``application`` logger has a file handler, then that file handler
      level is set. Otherwise, the console handler level is set.

    The logging configuration file ``formatters``, ``handlers`` and
    ``loggers`` sections are updated incrementally. For example, the
    ``conf/logging.yaml`` source distribution file defines the ``default``
    formatter ``format`` and ``datefmt``. If the  ``logging.yaml`` file in
    the current directory overrides the ``format`` but not the ``datefmt``,
    then the default ``datefmt`` is retained rather than unset. Thus, a custom
    logging configuration file need define only the settings which override
    the default configuration.

    By default, ``ERROR`` level messages are written to the  console.
    If the log file is set, then the default logger writes ``INFO`` level
    messages to a rotating log file.
    
    If the file handler is enabled, then this :meth:`qiutil.logging.configure`
    method ensures that the log file parent directory exists.

    Examples:

    - Write to the log:

      >>> from qiutil.logging import logger
      >>> logger(__name__).debug("Started the application...")

      or, in a class instance:

      >>> from qiutil.logging import logger
      >>> class MyApp(object):
      ...     def __init__(self):
      ...         self._logger = logger(__name__)
      ...     def start(self):
      ...         self._logger.debug("Started the application...")

    - Write debug messages to the file log:

      >>> import qiutil
      >>> qiutil.logging.configure(level='DEBUG')

    - Set the log file:

      >>> import qiutil
      >>> qiutil.logging.configure(filename='log/myapp.log')

    - Define your own logging configuration:

      >>> import qiutil
      >>> qiutil.logging.configure('/path/to/my/conf/logging.yaml')

    - Simplify the console log message format::

        ./logging.yaml:
        ---
        formatters:
          simple:
            format: '%(name)s - %(message)s'

        handlers:
          console:
            formatter: simple

    .. _YAML: http://www.yaml.org

    :param app: the application name
    :param cfg_file: the optional custom configuration YAML file
    :param opts: the Python ``logging.conf`` options, as well as the
        following short-cuts:
    :keyword filename: the log file path
    :keyword level: the file handler log level
    """
    # Load the configuration files.
    cfg = _load_config(app, cfg_file)

    # The options override the configuration files.
    if 'filename' in opts:
        fname = opts['filename']
        if fname:
            # Reset the log file.
            cfg['handlers']['file_handler']['filename'] = fname
        else:
            # Disable the file handler and retain only
            # the console handler.
            cfg['loggers'][app]['handlers'] = ['console']

    # The log level is set in both the logger and the handler,
    # and the more restrictive level applies. Therefore, set
    # the log level in both places.
    level = opts.pop('level', None)
    if level:
        cfg['loggers'][app]['level'] = level
        if 'file_handler' in cfg['loggers'][app]['handlers']:
            cfg['handlers']['file_handler']['level'] = level
        else:
            cfg['handlers']['console']['level'] = level

    # Add the other options, if any.
    qicollections.update(cfg, opts, recursive=True)

    # Ensure that the log file parent directory exists.
    if 'file_handler' in cfg['loggers'][app]['handlers']:
        log_file = cfg['handlers']['file_handler'].get('filename')
        if not log_file:
            raise ValueError("The logging configuration file handler is enabled"
                             " but the log file name is not set.")
        # Make the runtime path absolute for clarity.
        log_file = os.path.abspath(log_file)
        cfg['handlers']['file_handler']['filename'] = log_file
        # Make the log file parent directory, if necessary.
        log_dir = os.path.dirname(log_file)
        if log_dir and not os.path.exists(log_dir):
            os.makedirs(log_dir)

    # Configure the logger.
    logging.config.dictConfig(cfg)

    # Set the logger configured flag.
    setattr(logger, 'configured', True)


LOG_CFG_ENV_VAR = 'LOG_CONFIG'
"""The user-defined environment variable logging configuration file path."""

LOG_CFG_FILE = 'logging.yaml'
"""The optional current working directory logging configuration file name."""

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
"""The ``qiutil`` package directory."""

DEF_LOG_CFG = os.path.join(BASE_DIR, 'conf', LOG_CFG_FILE)
"""The default application logging configuration file path."""


def _load_config(app, cfg_file=None):
    """
    Loads the logger configuration files, as described in
    :meth:`qiutil.logging.configure`.

    :return: the logging configuration dictionary
    :raises ValueError: if the configuration file argument is specified
        but does not exist
    """
    # The base config file.
    config = _load_config_file(DEF_LOG_CFG)
    # Rename the application logger.
    loggers = config['loggers']
    loggers[app] = loggers.pop('application')
    
    # The custom configuration files.
    custom_cfg_files = _find_custom_config_files(cfg_file)
    # Load the custom configurations.
    custom_cfgs = (_load_config_file(f) for f in custom_cfg_files)
    # Update the base configuration.
    qicollections.update(config, *custom_cfgs, recursive=True)

    return config


def _find_custom_config_files(cfg_file):
    """
    Finds the custom logging configuration files, as described in
    :meth:`qiutil.logging.configure`.

    :param cfg_file: the custom configuration file argument
    :return: the custom configuration file list
    :raises ValueError: if the configuration file argument is specified
        but does not exist
    """
    # The config files list.
    config_files = []
    
    # The environment variable log configuration file.
    env_cfg_file = os.getenv(LOG_CFG_ENV_VAR, None)
    if env_cfg_file and os.path.exists(env_cfg_file):
        config_files.append(env_cfg_file)

    # The current directory log configuration file.
    if os.path.exists(LOG_CFG_FILE):
        config_files.append(LOG_CFG_FILE)

    # The argument log configuration file.
    if cfg_file:
        if os.path.exists(cfg_file):
            config_files.append(cfg_file)
        else:
            raise ValueError("Configuration file not found: %s" % cfg_file)

    return config_files


def _load_config_file(filename):
    """
    Loads the given logger configuration file.

    :param: filename: the log configuration file path
    :return: the parsed configuration parameter dictionary
    """
    with open(filename) as fs:
        return yaml.load(fs)
